from django.apps import AppConfig
from django.core.exceptions import ImproperlyConfigured


class CantoConfig(AppConfig):
    name = "canto"

    def ready(self):
        from django.conf import settings

        settings = settings._wrapped.__dict__

        settings.setdefault(
            "CANTO_OAUTH_URL", "https://oauth.canto.com:8443/oauth/api/oauth2/authorize"
        )
        settings.setdefault(
            "CANTO_OAUTH_TOKEN_URL",
            "https://oauth.canto.com:8443/oauth/api/oauth2/token",
        )
        settings.setdefault("CANTO_FILTER_SCHEMES", "")  # eg image or image|video

        if not settings.get("CANTO_API_URL"):
            raise ImproperlyConfigured("setting CANTO_API_URL is required.")
        if not settings.get("CANTO_APP_ID"):
            raise ImproperlyConfigured("setting CANTO_APP_ID is required.")
        if not settings.get("CANTO_APP_SECRET"):
            raise ImproperlyConfigured("setting CANTO_APP_SECRET is required.")



import math
from collections import OrderedDict
from datetime import timedelta
from json import JSONDecodeError
from logging import getLogger
from urllib.parse import urlencode

import requests

try:
    from django.utils.timezone import now
except ImportError:
    import datetime

    now = datetime.datetime.now

logger = getLogger(__name__)


class ApiException(Exception):
    pass


class ResourceNotFound(ApiException):
    pass


class AccessTokenRequired(ApiException):
    pass


class CantoClient:
    def __init__(
        self, api_url, app_id, app_secret, oauth_url, oauth_token_url, access_token
    ):
        self.api_url = api_url
        self.app_id = app_id
        self.app_secret = app_secret
        self.oauth_url = oauth_url
        self.oauth_token_url = oauth_token_url
        self.access_token = access_token

    def create_access_token(self, code):
        params = {
            "code": code,
            "app_secret": self.app_secret,
            "grant_type": "authorization_code",
            "app_id": self.app_id,
        }
        url = "{}?{}".format(self.oauth_token_url, urlencode(params))

        request_time = now()
        response = requests.post(url)

        if response.ok:
            data = response.json()
            access_token = data["accessToken"]  # valid for one month

            expires_in_seconds = int(data["expiresIn"])
            valid_until = request_time + timedelta(seconds=expires_in_seconds)
            refresh_token = data["refreshToken"]  # valid for one year
            return access_token, valid_until, refresh_token
        else:
            try:
                data = response.json()
                error_code = data["error"]
                error_description = data["error_description"]
            except JSONDecodeError:
                error_code = ""
                error_description = response.content

            logger.error(
                "An error occurred getting an access token: %s %s",
                error_code,
                error_description,
            )
            raise ApiException("Error getting access token")

    def refresh_access_token(self, refresh_token):
        params = {
            "app_secret": self.app_secret,
            "refresh_token": refresh_token,
            "grant_type": "refresh_token",
            "app_id": self.app_id,
        }
        url = "{}?{}".format(self.oauth_token_url, urlencode(params))

        request_time = now()
        response = requests.post(url)

        if response.ok:
            data = response.json()
            access_token = data["accessToken"]  # valid for one month
            expires_in_seconds = int(data["expiresIn"])
            valid_until = request_time + timedelta(seconds=expires_in_seconds)
            refresh_token = data["refreshToken"]  # valid for one year
            return access_token, valid_until, refresh_token
        else:
            try:
                data = response.json()
                error_code = data["error"]
                error_description = data["error_description"]
            except JSONDecodeError:
                error_code = ""
                error_description = response.content

            logger.error(
                "An error occurred getting an access token: %s %s",
                error_code,
                error_description,
            )
            raise ApiException("Error getting access token")

    def _authenticated_request(self, url, params=None, method='GET', **kwargs):
        if not self.access_token:
            raise AccessTokenRequired("an access token is required")

        if not url.startswith(self.api_url):
            url = self.api_url + url

        if params:
            url += "?" + urlencode(params)

        method = method.lower()
        if method not in ['get', 'post']:
            raise ApiException(
                'method should be one of "get" or "post". Got "{}".'.format(method)
            )

        response = requests.request(
            method,
            url,
            headers={
                "Authorization": "Bearer {}".format(self.access_token),
                "User-Agent": self.app_id,
            },
            **kwargs,
        )

        if response.ok:
            return response
        elif response.status_code == 404:
            raise ResourceNotFound()
        else:
            raise ApiException(
                "response status code was {}".format(response.status_code)
            )

    def test_connection(self):
        return self._authenticated_request("/api/v1/").json()

    def get_oauth_url(self, state, redirect_uri):
        params = {
            "response_type": "code",
            "app_id": self.app_id,
            "redirect_uri": redirect_uri,
            "state": state,
        }
        return "{}?{}".format(self.oauth_url, urlencode(params))

    def get_album(
        self,
        album_id,
        start,
        paginate_by,
        scheme,
        sort_by="time",
        sort_direction="descending",
    ):
        # start = max(page - 1, 0) * paginate_by
        params = {
            "limit": paginate_by,
            "start": start,
            "scheme": scheme,
            "sortBy": sort_by,
            "sortDirection": sort_direction,
        }

        album_data = self._authenticated_request(
            "/api/v1/album/{}".format(album_id), params
        ).json()

        return album_data

    def get_search_results(
        self,
        query,
        start,
        paginate_by,
        scheme,
        sort_by="time",
        sort_direction="descending",
    ):
        params = {
            "limit": paginate_by,
            "start": start,
            "keyword": query,
            "scheme": scheme,
            "sortBy": sort_by,
            "sortDirection": sort_direction,
        }

        search_results = self._authenticated_request("/api/v1/search", params).json()

        return search_results

    def get_image(self, image_id):
        image_data = self._authenticated_request(
            "/api/v1/image/{}".format(image_id)
        ).json()

        return image_data

    def get_binary(self, url):
        assert url.startswith(self.api_url + "/api_binary/")
        response = self._authenticated_request(url, allow_redirects=True)
        return response.content

    def get_public_url_for_binary(self, url):
        assert url.startswith(self.api_url + "/api_binary/"), (
            url,
            self.api_url + "/api_binary/",
        )
        response = self._authenticated_request(url, allow_redirects=False)

        if not response.status_code == 302:
            raise ApiException(
                "Unexpected response code {}, expected a redirect".format(
                    response.status_code
                )
            )

        return response.headers["Location"]

    def get_tree(self):
        return self._authenticated_request(
            "/api/v1/tree", {"sortBy": "name", "sortDirection": "ascending"}
        ).json()

    def create_folder(self, folder_name, parent_folder_id=None, description=None):
        if parent_folder_id:
            url = '/api/v1/folder/{}/{}'.format(parent_folder_id, folder_name)
        else:
            url = '/api/v1/folder/{}'.format(folder_name)

        data = {}
        if description:
            data["Description"] = '{}'.format(description)

        return self._authenticated_request(url, method='post', data=data).json()

    def create_album(self, album_name, parent_folder_id=None, description=None):
        if parent_folder_id:
            url = '/api/v1/album/{}/{}'.format(parent_folder_id, album_name)
        else:
            url = '/api/v1/album/{}'.format(album_name)

        data = {}
        if description:
            data["Description"] = '{}'.format(description)

        return self._authenticated_request(url, method='post', data=data).json()

    def get_upload_setting(self):
        url = '/api/v1/upload/setting'
        return self._authenticated_request(url).json()

    def upload_file(self, file, album_id, image_id="", scheme=""):
        setting = self.get_upload_setting()
        url = setting.get("url", None)
        if not url:
            raise ApiException("Failed to get canto upload setting")

        # we use an ordered dict, since the file must be the last parameter
        # see https://www.canto.com/api/?api=com#Upload-file
        data = {
            "key": setting["key"],
            "acl": setting["acl"],
            "AWSAccessKeyId": setting["AWSAccessKeyId"],
            "Policy": setting["Policy"],
            "Signature": setting["Signature"],

            "x-amz-meta-file_name": "${filename}",
            "x-amz-meta-tag": "",
            "x-amz-meta-scheme": scheme,  # set original scheme if updating
            "x-amz-meta-id": image_id,  # set original id if updating
            "x-amz-meta-album_id": album_id,
        }
        files = {'file': file}

        response = requests.post(
            url,
            data=data,
            files=files
        )

        if response.ok:
            return response
        else:
            raise ApiException(
                "response status code was {}".format(response.status_code)
            )




from django.contrib.admin.widgets import ForeignKeyRawIdWidget
from django.urls import reverse
from django.utils.safestring import mark_safe
from django.utils.translation import ugettext as _


class CantoImageWidget(ForeignKeyRawIdWidget):
    def get_context(self, name, value, attrs):
        context = super().get_context(name, value, attrs)
        rel_to = self.rel.model
        if rel_to in self.admin_site._registry:
            # The related object is registered with the same AdminSite
            related_url = reverse("canto:library")
            context["related_url"] = mark_safe(related_url)
            context["link_title"] = _("Lookup")
        return context

    def __init__(self, rel, admin_site, attrs=None, using=None):
        attrs = attrs or {}
        attrs["data-canto-image"] = "true"
        super().__init__(rel, admin_site, attrs, using)



from datetime import timedelta

import mock
from django.contrib.auth import get_user_model
from django.contrib.auth.models import Permission
from django.test import override_settings
from django.urls import reverse
from django.utils import timezone
from django.utils.html import escape
from django_dynamic_fixture import G
from django_webtest import WebTest

from canto.services import _get_oauth_state, _get_canto_settings, get_canto_client


class FakeResponse:
    def __init__(self, data, ok=True) -> None:
        super().__init__()
        self.data = data
        self.ok = ok

    def json(self):
        return self.data


@override_settings(
    CANTO_API_URL="https://example.canto.com",
    CANTO_APP_ID="XXX",
    CANTO_APP_SECRET="ZZZ",
)
class CantoViewTest(WebTest):
    def setUp(self):
        self.user = G(get_user_model())
        self.user.user_permissions.add(
            Permission.objects.get(codename="browse_library")
        )
        self.user.user_permissions.add(
            Permission.objects.get(codename="change_cantosettings")
        )

    def test_settings_default_to_not_connected(self):
        response = self.app.get(reverse("canto:settings"), user=self.user)
        self.assertContains(response, "You are not connected")
        self.assertContains(
            response,
            escape(
                get_canto_client().get_oauth_url(
                    _get_oauth_state(self.user), response.request.url
                )
            ),
        )

    def test_settings_permissions(self):
        unauthorized_user = G(get_user_model())
        self.app.get(reverse("canto:settings"), user=unauthorized_user, status=403)
        self.app.post(reverse("canto:disconnect"), user=unauthorized_user, status=403)
        self.app.post(
            reverse("canto:refresh-token"), user=unauthorized_user, status=403
        )

    def test_browse_permissions(self):
        unauthorized_user = G(get_user_model())
        self.app.get(reverse("canto:library"), user=unauthorized_user, status=403)
        self.app.get(reverse("canto:tree-json"), user=unauthorized_user, status=403)
        self.app.get(
            reverse("canto:search-json", kwargs={"query": "something"}),
            user=unauthorized_user,
            status=403,
        )
        self.app.get(
            reverse("canto:album-json", kwargs={"album_id": "123"}),
            user=unauthorized_user,
            status=403,
        )
        self.app.get(
            reverse("canto:binary", kwargs={"url": "foo"}),
            user=unauthorized_user,
            status=403,
        )

    def test_oauth_confirmation_view(self):
        oauth_state = _get_oauth_state(self.user)
        settings_page = self.app.get(
            reverse("canto:settings") + "?code=CANTO_CODE&state=" + oauth_state,
            user=self.user,
        )
        self.assertContains(settings_page, "Please confirm the connection to canto")

        canto_response = FakeResponse(
            {
                "accessToken": "i-grant-access",
                "state": oauth_state,
                "expiresIn": 3600,
                "refreshToken": "feel-refreshed",
            }
        )
        now = timezone.now()

        with mock.patch("canto.client.requests.post", return_value=canto_response):
            with mock.patch("canto.client.now", return_value=now):
                confirm_response = settings_page.form.submit()

        self.assertContains(confirm_response.follow(), "You are connected to canto")

        canto_settings = _get_canto_settings()
        self.assertEqual(canto_settings.access_token, "i-grant-access")
        self.assertEqual(canto_settings.refresh_token, "feel-refreshed")
        self.assertEqual(
            canto_settings.token_valid_until, now + timedelta(seconds=3600)
        )

    def test_oauth_disconnect_view(self):
        canto_settings = _get_canto_settings()
        canto_settings.refresh_token = "refresh"
        canto_settings.access_token = "access"
        canto_settings.token_valid_until = timezone.now() + timedelta(seconds=3600)
        canto_settings.save()

        response = self.app.get(reverse("canto:settings"), user=self.user)
        self.assertContains(response, "You are connected to canto")

        response = response.forms["canto-disconnect"].submit()
        self.assertContains(response.follow(), "You are not connected")



import hashlib

from django.conf import settings

from canto.client import CantoClient
from canto.models import CantoSettings


def _get_canto_settings():
    return CantoSettings.objects.get_or_create()[0]


def get_canto_client(canto_settings: CantoSettings = None):
    canto_settings = canto_settings or _get_canto_settings()
    return CantoClient(
        api_url=settings.CANTO_API_URL,
        app_id=settings.CANTO_APP_ID,
        app_secret=settings.CANTO_APP_SECRET,
        oauth_url=settings.CANTO_OAUTH_URL,
        oauth_token_url=settings.CANTO_OAUTH_TOKEN_URL,
        access_token=canto_settings.access_token,
    )


def get_and_save_access_token(code, state, expected_state):
    if state != expected_state:
        # we ensure that the requested state matches the response state
        # https://auth0.com/docs/protocols/oauth2/oauth-state
        raise Exception("invalid state")

    access_token, valid_until, refresh_token = get_canto_client().create_access_token(
        code
    )

    CantoSettings.objects.update_or_create(
        defaults={
            "access_token": access_token,
            "refresh_token": refresh_token,
            "token_valid_until": valid_until,
        }
    )


def refresh_and_save_access_token():
    canto_settings = _get_canto_settings()
    assert canto_settings.refresh_token
    access_token, valid_until, refresh_token = get_canto_client().refresh_access_token(
        canto_settings.refresh_token
    )
    canto_settings.access_token = access_token
    canto_settings.refresh_token = refresh_token
    canto_settings.token_valid_until = valid_until
    canto_settings.save()


def _get_oauth_state(user):
    """
    Local state that is passed to the oauth endpoint and back to ensure the response matches
    the request.
    """
    assert user.is_authenticated
    canto_settings = _get_canto_settings()
    return hashlib.sha512(
        "{}-{}".format(canto_settings.last_modified_at, user.pk).encode("utf-8")
    ).hexdigest()


def disconnect_canto():
    try:
        canto_settings = _get_canto_settings()
    except CantoSettings.DoesNotExist:
        return

    canto_settings.refresh_token = ""
    canto_settings.access_token = ""
    canto_settings.token_valid_until = None
    canto_settings.save()



from django.conf.urls import url

from canto.views import (
    CantoSettingsView,
    refresh_token,
    disconnect,
    canto_binary_view,
    CantoLibraryView,
    CantoAlbumJsonView,
    CantoTreeJsonView,
    CantoSearchJsonView,
)

app_name = "canto"
urlpatterns = [
    url(r"^canto/settings/$", CantoSettingsView.as_view(), name="settings"),
    url(r"^canto/refresh/$", refresh_token, name="refresh-token"),
    url(r"^canto/disconnect/$", disconnect, name="disconnect"),
    url(r"^canto/library/$", CantoLibraryView.as_view(), name="library"),
    url(r"^canto/tree.json$", CantoTreeJsonView.as_view(), name="tree-json"),
    url(
        r"^canto/search/(?P<query>.+).json$",
        CantoSearchJsonView.as_view(),
        name="search-json",
    ),
    url(
        r"^canto/album/(?P<album_id>.+).json$",
        CantoAlbumJsonView.as_view(),
        name="album-json",
    ),
    url(r"^canto/binary/(?P<url>.+)$", canto_binary_view, name="binary"),
]



from django.core.management.base import BaseCommand

from ...services import refresh_and_save_access_token


class Command(BaseCommand):
    def handle(self, *args, **options):
        refresh_and_save_access_token()



from django.db import models
from django.utils import timezone


class CantoSettings(models.Model):
    class Meta:
        permissions = (("browse_library", "Can browse the canto library"),)

    access_token = models.CharField(max_length=255)
    refresh_token = models.CharField(max_length=255)

    token_valid_until = models.DateTimeField(null=True)

    last_modified_at = models.DateTimeField(auto_now=True)

    def is_connected(self):
        return self.access_token and self.token_valid_until > timezone.now()



# Generated by Django 2.1.1 on 2018-10-23 09:23

from django.db import migrations, models


class Migration(migrations.Migration):

    initial = True

    dependencies = []

    operations = [
        migrations.CreateModel(
            name="CantoSettings",
            fields=[
                (
                    "id",
                    models.AutoField(
                        auto_created=True,
                        primary_key=True,
                        serialize=False,
                        verbose_name="ID",
                    ),
                ),
                ("access_token", models.CharField(max_length=255)),
                ("refresh_token", models.CharField(max_length=255)),
                ("token_valid_until", models.DateTimeField(null=True)),
                ("last_modified_at", models.DateTimeField(auto_now=True)),
            ],
            options={
                "permissions": (("browse_library", "Can browse the canto library"),),
            },
        )
    ]



from logging import getLogger

from django.conf import settings
from django.contrib.auth.decorators import permission_required
from django.contrib.auth.mixins import PermissionRequiredMixin
from django.contrib.messages import success
from django.http import HttpResponseRedirect, JsonResponse
from django.urls import reverse_lazy
from django.utils.translation import ugettext as _
from django.views import View
from django.views.decorators.cache import cache_control
from django.views.decorators.http import require_POST
from django.views.generic import FormView, TemplateView

from canto.services import (
    get_and_save_access_token,
    refresh_and_save_access_token,
    _get_oauth_state,
    disconnect_canto,
    _get_canto_settings,
    get_canto_client,
)
from .forms import CantoSettingsForm

logger = getLogger(__name__)


@permission_required("canto.change_cantosettings", raise_exception=True)
@require_POST
def refresh_token(request, success_url=reverse_lazy("canto:settings")):
    refresh_and_save_access_token()
    success(request, _("Your canto token was refreshed."))
    return HttpResponseRedirect(success_url)


@permission_required("canto.change_cantosettings", raise_exception=True)
@require_POST
def disconnect(request, success_url=reverse_lazy("canto:settings")):
    disconnect_canto()
    success(request, _("Canto was disconnected."))
    return HttpResponseRedirect(success_url)


class CantoSettingsView(PermissionRequiredMixin, FormView):
    permission_required = "canto.change_cantosettings"
    form_class = CantoSettingsForm
    template_name = "canto/settings.html"

    success_url = reverse_lazy("canto:settings")

    def dispatch(self, request, *args, **kwargs):
        self.canto_settings = _get_canto_settings()
        self.canto_client = get_canto_client(self.canto_settings)
        return super().dispatch(request, *args, **kwargs)

    def get_context_data(self, **kwargs):
        context = super().get_context_data(**kwargs)

        context["title"] = _("Canto settings")
        context["settings"] = self.canto_settings
        context["oauth_url"] = self.canto_client.get_oauth_url(
            _get_oauth_state(self.request.user),
            self.request.build_absolute_uri(self.request.path),
        )
        context["oauth_error_code"], context[
            "oauth_error_message"
        ] = self.get_oauth_error()

        return context

    def get_form_kwargs(self):
        kwargs = super().get_form_kwargs()
        kwargs["data"] = {
            "code": self.request.GET.get("code", ""),
            "state": self.request.GET.get("state", ""),
        }
        return kwargs

    def form_valid(self, form):
        get_and_save_access_token(
            form.cleaned_data["code"],
            state=form.cleaned_data["state"],
            expected_state=_get_oauth_state(self.request.user),
        )

        success(self.request, "Connected to canto!")
        return super().form_valid(form)

    def get_oauth_error(self):
        error_code = self.request.GET.get("error", "")
        error_message = self.request.GET.get("error_description", "")

        if error_code:
            logger.error(
                "An oauth error has occurred: %s %s", error_code, error_message
            )

        return error_code, error_message


class CantoLibraryView(PermissionRequiredMixin, TemplateView):
    permission_required = "canto.browse_library"
    template_name = "canto/library.html"
    title = _("Canto library")
    paginate_by = 100

    def get_context_data(self, **kwargs):
        context = {"title": self.title}
        context.update(kwargs)
        return super().get_context_data(**context)


@permission_required("canto.browse_library", raise_exception=True)
@cache_control(max_age=300)
def canto_binary_view(request, url):
    # double slashed may be stripped by the web server
    if url.startswith('https:/') and not url.startswith('https://'):
        url = 'https://' + url[len('https:/'):]
    public_url = get_canto_client().get_public_url_for_binary(url)
    return HttpResponseRedirect(public_url)


class CantoTreeJsonView(PermissionRequiredMixin, View):
    permission_required = "canto.browse_library"

    def get(self, request, **kwargs):
        results = get_canto_client().get_tree()
        return JsonResponse(results)


class CantoAlbumJsonView(PermissionRequiredMixin, View):
    permission_required = "canto.browse_library"
    paginate_by = 100
    album_id = None

    def get(self, request, **kwargs):
        try:
            start = int(request.GET.get("start", 0))
        except ValueError:
            start = 0

        results = get_canto_client().get_album(
            self.kwargs["album_id"],
            start,
            self.paginate_by,
            settings.CANTO_FILTER_SCHEMES,
        )

        return JsonResponse(results)


class CantoSearchJsonView(PermissionRequiredMixin, View):
    permission_required = "canto.browse_library"
    paginate_by = 100

    def get(self, request, **kwargs):
        try:
            start = int(request.GET.get("start", 0))
        except ValueError:
            start = 0

        results = get_canto_client().get_search_results(
            kwargs["query"], start, self.paginate_by, settings.CANTO_FILTER_SCHEMES
        )

        return JsonResponse(results)
